#include<stdio.h>
#include<stdlib.h>
#include <regex.h>
#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_http.h>
#include <ngx_md5.h>
#include <ngx_parse.h>
#include <ngx_palloc.h>
#include <ngx_string.h>

// MD5 produces a 128 bit / 8 = 16 byte hash
#define HASH_SIZE 16
#define X_OLIVER_BALANCED "X-Oliver-Balanced"

// NODE #######################################################################
// Initialized when a configuration is loaded.  Represents a server.
typedef struct {
	u_char hash[HASH_SIZE];
	struct sockaddr *sockaddr;
	socklen_t socklen;
	ngx_str_t name;
} node;

node* node_new(ngx_str_t name);
node* node_init(node* n,ngx_str_t);
static int node_cmp(const void *va, const void *vb);
void node_free(node *n);

// REQUEST NODE ###############################################################
// Initialized at the beginning of a request. Encapsulates information about
// whether or not this node has been tried.
typedef struct {
	node *node;
	u_char tried;
} request_node;

// NODE RING ##################################################################
// Initialized when a configuration is loaded.  Represents a ring of servers.
typedef struct {
	node * nodes;
	unsigned int n_nodes;
	u_char n_replicas;
	regex_t regex;
	ngx_str_t alias;
	int local_node_index;
} ring;

ring* ring_new(
	int n_replicas,
	ngx_str_t alias,
	int n_nodes,
	ngx_conf_t *cf,
	ngx_http_upstream_srv_conf_t *us);
void ring_free(ring *r);
int ring_assign(ring *r,ngx_str_t);
int ring_assign_random(ring *r);

// REQUEST NODE RING ###########################################################

// Initialized at the beginning of a request.  Encapsulates information about
// which nodes have been tried already
typedef struct {
	ring *ring;
	request_node *rnodes;
	unsigned int point;
	unsigned int loop_size;
	unsigned int pos;
	unsigned int tries;
	ngx_http_request_t *request;
	ngx_peer_connection_t *peer_connection;
	u_char balanced_elsewhere;
	ngx_str_t key;
	ngx_str_t balanced_header;
} request_ring;

request_ring* request_ring_new(
		ring* ring,
		ngx_http_request_t *r);
int request_ring_assign(request_ring *rr);
void request_ring_free(request_ring *rr);
unsigned int request_ring_current_pos(request_ring*);
void request_ring_advance(request_ring*);
u_char request_ring_exhausted(request_ring*);
u_char request_ring_has_alias(request_ring*);
unsigned int request_ring_candidate(request_ring*);

// UTIL #######################################################################

static void md5(u_char *str, size_t len,u_char *md5_buf);
//static char* ngx_str_to_c_str(ngx_str_t str,ngx_http_request_t *request);

// NGINX ######################################################################
/*
 * 1) Enabling Configuration Directive calls a user-defined registration function
 * 2) User-defined registration function defines legal directives (e.g weight),
 *    and registers a user-defined upstream initialization function
 * 3) User-defined upstream initialization function
 * 		- resolves server names to ip addresses
 * 		- allocates space for sockets
 * 		- registers a user-defined peer initialization function
 * 	4) User-defined peer initialization function is called once per request, and
 * 	   sets up data structures used by the load balancing function
 * 	5) User-defined load balancing functions is called one or more times per
 * 	   request, depending on if requests fail
 * 	6) User-defined peer release function is called one or more times per request,
 * 	   and updates statistics (or whatever)
 */

typedef struct {
	unsigned int n_replicas;
	ngx_str_t alias;
} consistent_hash_conf;

static void *ngx_http_upstream_consistent_hash_create_conf(ngx_conf_t *);

// User-defined registration function defines legal directives (e.g weight),
// and registers a user-defined upstream initialization function
static char * ngx_http_upstream_consistent_hash(
		ngx_conf_t*,
        ngx_command_t*,
        void*);

// User-defined upstream initialization function
// 	- resolves server names to ip addresses
// 	- allocates space for sockets
//	- registers a peer initialization function
static ngx_int_t ngx_http_upstream_init_consistent_hash(
		ngx_conf_t *,
        ngx_http_upstream_srv_conf_t*);

// Called once per request.  Sets up data structures to be used by the load
// balancing and peer release functions
static ngx_int_t ngx_http_upstream_init_consistent_hash_peer(
        ngx_http_request_t*,
        ngx_http_upstream_srv_conf_t*);

// Load balancing algorithm
static ngx_int_t ngx_http_upstream_get_consistent_hash_peer(
        ngx_peer_connection_t*,
        void*);

// Called one or more times per request.  Updates statistics about completed
// requests, successful or not.
static void ngx_http_upstream_free_consistent_hash_peer(
		ngx_peer_connection_t *pc,
		void *data,
		ngx_uint_t state);

// NODE IMPL ##################################################################

node* node_new(ngx_str_t name) {
	node* n = (node*)malloc(sizeof(node));
	return node_init(n,name);
}

node* node_init(node* n,ngx_str_t name) {
	n->name = name;
	md5(name.data,name.len,n->hash);
	return n;
}

static int node_cmp(const void *va, const void *vb) {
	u_char *a =  ((node*)va)->hash;
	u_char *b =  ((node*)vb)->hash;
	int pos;
	for(pos = 0; pos < HASH_SIZE; pos ++) {
		if(a[pos] < b[pos]) { return -1; }
		if(a[pos] > b[pos]) { return 1; }
		continue;
	}
	return 0;
}

void node_free(node *n) {
	free(n);
}

// NODE RING IMPL #############################################################

ring* ring_new(
	int n_replicas,
	ngx_str_t alias,
	int n_nodes,ngx_conf_t *cf,
	ngx_http_upstream_srv_conf_t *us) {

	ngx_http_upstream_server_t      *server;
	server = us->servers->elts;

	if(n_replicas < 1 || n_nodes < 1) { return NULL; }
	if(n_replicas > n_nodes) {
		n_replicas = n_nodes;
	}

	node *nodes = (node*)ngx_pcalloc(cf->pool,sizeof(node) * n_nodes);
	ring *r = (ring*)ngx_pcalloc(cf->pool,sizeof(ring));
	r->n_replicas = n_replicas;
	r->n_nodes = n_nodes;
	r->alias = alias;
	unsigned int i,j,k = 0;
	for (i = 0; i < us->servers->nelts; i++) {
		for (j = 0; j < server[i].naddrs; j++) {
			nodes[k].sockaddr = server[i].addrs[j].sockaddr;
			nodes[k].socklen = server[i].addrs[j].socklen;
			node_init(
				&nodes[k],
				k == 0 && alias.len ? alias : server[i].addrs[j].name);
			k++;
		}
	}

	qsort(nodes,n_nodes,sizeof(node),node_cmp);
	r->nodes = nodes;
	regcomp(&r->regex,"/data/[^/[:space:]?]+",REG_EXTENDED);

	r->local_node_index = -1;
	if(!alias.len) { return r; }

	/* BEGIN DEBUG/TEST CODE */
	unsigned int m = 0;
	for(m = 0; m < r->n_nodes; m++) {
		printf("%s\n",r->nodes[m].name.data);
	}
	printf("==============================\n");
	ngx_str_t Oliver = ngx_string("Oliver");
	ngx_str_t Rebekah = ngx_string("Rebekah");
	ngx_str_t Tucker = ngx_string("Tucker");
	ngx_str_t John = ngx_string("John");
	printf("%i\n",ring_assign(r,Oliver));
	printf("%i\n",ring_assign(r,Rebekah));
	printf("%i\n",ring_assign(r,Tucker));
	printf("%i\n",ring_assign(r,John));
	printf("==============================\n");
	printf("%s\n",Oliver.data);
	printf("%s\n",Rebekah.data);
	printf("%s\n",Tucker.data);
	printf("%s\n",John.data);
	/* END DEBUG/TEST CODE */

	unsigned int l = 0;
	// record the index of the local node
	for(l = 0; l < r->n_nodes; l++) {
		if(0 == ngx_strcmp(alias.data,r->nodes[l].name.data)) {
			r->local_node_index = l;
			break;
		}
	}
	/* BEGIN DEBUG/TEST CODE */
	printf("==============================\n");
	printf("%i\n",r->local_node_index);
	/* END DEBUG/TEST CODE */
	return r;
}

void ring_free(ring *r) {
	free(r->nodes);
	free(r);
}

int ring_assign(ring* r,ngx_str_t uri) {
	unsigned int i;
	node* n = node_new(uri);
	for(i = 0; i < r->n_nodes; i++) {
		if(node_cmp((void *)n,(void *)&(r->nodes[i % r->n_nodes])) <= 0) {
			break;
		}
	}
	node_free(n);
	return i % r->n_nodes;
}

int ring_assign_random(ring *r) {
	return rand() % r->n_nodes;
}

// REQUEST NODE RING IMPL #####################################################

u_char request_ring_balanced_elsewhere(request_ring* rr) {
	ngx_list_t headers = rr->request->headers_in.headers;
	ngx_table_elt_t *data;
	ngx_list_part_t *part;
	unsigned int i;

	part = &headers.part;
	data = part->elts;

	for (i = 0 ;; i++) {
		if (i >= part->nelts) {
			if (part->next == NULL) {
				break;
			}

			part = part->next;
			data = part->elts;
			i = 0;
		}

		if(0 == ngx_strcmp(data[i].key.data,X_OLIVER_BALANCED)
		   /*&& NULL != strchr((const char*)data[i].value.data,',')*/) {
			rr->balanced_header.data = data[i].value.data;
			rr->balanced_header.len = data[i].value.len;
			return 1;
		}
	}

	rr->balanced_header.data = (u_char*)"\0";
	rr->balanced_header.len = 0;
	return 0;
}

request_ring* request_ring_new(
		ring* ring,
		ngx_http_request_t *r) {

	regmatch_t match[1];
	ngx_str_t key;
	unsigned int i,reti;
	request_node *nodes;
	request_ring *rr;

	nodes = (request_node*)ngx_pcalloc(
			r->pool,sizeof(request_node) * ring->n_nodes);
	for(i = 0; i < ring->n_nodes; i++) {
		nodes[i].node = &ring->nodes[i];
		nodes[i].tried = 0;
	}

	rr = (request_ring*)ngx_pcalloc(r->pool,sizeof(request_ring));
	rr->request = r;
	rr->ring = ring;
	rr->rnodes = nodes;

	// If I have a local node, and this request has been balanced elsewhere,
	// then process locally, and don't do retries

	if(rr->ring->local_node_index > -1
	   && request_ring_balanced_elsewhere(rr)) {
		rr->point = rr->ring->local_node_index;
		rr->loop_size = 1;
		//rr->tries = rr->ring->n_replicas - 1;

		/*
		ngx_log_debug2(
			NGX_LOG_DEBUG_HTTP,
			rr->request->connection->log,
			0,
			"[OLIVER] BALANCED ELSEWHERE ID : %s, ASSIGNEE: %s",
			rr->request->uri.data,
			rr->ring->nodes[rr->point].name.data);
		*/
		rr->key.data = r->uri.data;
		rr->key.len = r->uri.len;
		rr->balanced_elsewhere = 1;

		return rr;
	}

	rr->balanced_elsewhere = 0;

	reti = regexec(
		&ring->regex,
		//ngx_str_to_c_str(r->uri,r),
		(const char*)r->uri.data,
		1,
		match,
		0);

	if(reti == 0) {
		// this is a request for a key
		// KLUDGE: The hardcoded "6" here is to account for the "/data/" portion
		// of the URI.  I should come up with a better way to handle this, so
		// if the uri changes in the tornado app, this still behaves correctly
		key.data = r->uri.data + (match[0].rm_so + 6);
		key.len = match[0].rm_eo - match[0].rm_so - 6;
		rr->key.data = key.data;
		rr->key.len = key.len;
		rr->point = ring_assign(ring,key);
		rr->loop_size = ring->n_replicas;

		/*
		ngx_log_debug2(
			NGX_LOG_DEBUG_HTTP,
			rr->request->connection->log,
			0,
			"[OLIVER] ASSIGNING ID : %s TO : %s",
			key.data,
			rr->ring->nodes[rr->point].name.data);
		*/



		/*
		// If I have a local node that's capable of serving the request, try
		// myself first
		if(rr->ring->local_node_index > -1) {
			int p =
				(rr->ring->local_node_index - rr->point) % rr->ring->n_nodes;
			if(p < rr->ring->n_replicas) {
				 rr->pos = p;
				 return rr;
			}
		}
		*/
	} else {
		// This is a request for something other than a key.  It should be
		// routed to the local node, if available
		rr->key.data = r->uri.data;
		rr->key.len = r->uri.len;
		if(rr->ring->local_node_index > -1) {
			rr->point = rr->ring->local_node_index;
			rr->loop_size = 1;

			//rr->tries = rr->ring->n_replicas - 1;

		} else {
			// there's no local node, so assign to a random node
			rr->point = ring_assign_random(ring);
			rr->loop_size = ring->n_nodes;
		}
	}

	rr->pos = rand() % rr->loop_size;
	return rr;
}

unsigned int request_ring_current_pos(request_ring* rr) {
	return (rr->point + (rr->pos % rr->loop_size)) % rr->ring->n_nodes;
}

void request_ring_advance(request_ring* rr) {
	rr->pos++;
}

u_char request_ring_exhausted(request_ring* rr) {
	return rr->tries == rr->loop_size ? 1 : 0;
}

u_char request_ring_has_alias(request_ring* rr) {
	return rr->ring->alias.len > 0;
}

unsigned int request_ring_candidate(request_ring* rr) {
	unsigned int pos = request_ring_current_pos(rr);
	rr->tries++;
	rr->rnodes[pos].tried = 1;
	return pos;
}

int request_ring_assign(request_ring *rr) {
	if(request_ring_exhausted(rr)) {
		return -1;
	}

	unsigned int c = request_ring_candidate(rr);
	request_ring_advance(rr);

	return c;
}

void request_ring_free(request_ring *rr) {
	free(rr->rnodes);
	free(rr);
}

// UTIL IMPL ##################################################################

static void md5(u_char *str, size_t len,u_char *md5_buf) {
    ngx_md5_t  md5;
    ngx_md5_init(&md5);
    ngx_md5_update(&md5, str, len);
    ngx_md5_final(md5_buf, &md5);
}
/*
static char* ngx_str_to_c_str(ngx_str_t str,ngx_http_request_t *request) {
	char *s = ngx_pcalloc(request->pool,str.len + 1);
	memcpy(s,str.data,str.len);
	s[str.len + 1] = '\0';
	return s;
}
*/
// NGINX IMPL #################################################################

static ngx_command_t  ngx_http_upstream_consistent_hash_commands[] = {

    {   ngx_string("consistent_hash"),
        NGX_HTTP_UPS_CONF|NGX_CONF_TAKE12,
        ngx_http_upstream_consistent_hash,
        0,
        0,
        NULL },

    ngx_null_command
};

static ngx_http_module_t  ngx_http_upstream_consistent_hash_module_ctx = {
    NULL,                                          /* preconfiguration */
    NULL,                                          /* postconfiguration */
    NULL,                                          /* create main configuration */
    NULL,                                          /* init main configuration */
    ngx_http_upstream_consistent_hash_create_conf, /* create server configuration */
    NULL,                                          /* merge server configuration */
    NULL,                                          /* create location configuration */
    NULL                                           /* merge location configuration */
};

ngx_module_t  ngx_http_upstream_consistent_hash_module = {
    NGX_MODULE_V1,
    &ngx_http_upstream_consistent_hash_module_ctx, /* module context */
    ngx_http_upstream_consistent_hash_commands,    /* module directives */
    NGX_HTTP_MODULE,                               /* module type */
    NULL,                                          /* init master */
    NULL,                                          /* init module */
    NULL,                                          /* init process */
    NULL,                                          /* init thread */
    NULL,                                          /* exit thread */
    NULL,                                          /* exit process */
    NULL,                                          /* exit master */
    NGX_MODULE_V1_PADDING
};

static void * ngx_http_upstream_consistent_hash_create_conf(ngx_conf_t *cf) {
	consistent_hash_conf  *conf;
    conf = ngx_pcalloc(cf->pool,sizeof(consistent_hash_conf));
    if (conf == NULL) {
        return NULL;
    }
    conf->n_replicas = 1;
    conf->alias.data = NULL;
    conf->alias.len = 0;
    return conf;
}

static char * ngx_http_upstream_consistent_hash(
		ngx_conf_t *cf, ngx_command_t *cmd, void *conf) {

	srand (time(NULL));
	ngx_str_t *arg;
	consistent_hash_conf * chcf;
	arg = cf->args->elts;
    ngx_http_upstream_srv_conf_t  *uscf;
    uscf = ngx_http_conf_get_module_srv_conf(
    		cf, ngx_http_upstream_module);
    chcf = ngx_http_conf_upstream_srv_conf(
    		uscf,ngx_http_upstream_consistent_hash_module);
    chcf->n_replicas = ngx_atoi(arg[1].data,arg[1].len);
    chcf->alias = arg[2];
    if(NULL == uscf) {
    	return NGX_CONF_ERROR;
    }
    uscf->peer.init_upstream = ngx_http_upstream_init_consistent_hash;
    uscf->flags = NGX_HTTP_UPSTREAM_CREATE;
    return NGX_CONF_OK;
}

static ngx_int_t ngx_http_upstream_init_consistent_hash(
		ngx_conf_t *cf,ngx_http_upstream_srv_conf_t *us) {

	ngx_uint_t                      i,n;
	ngx_http_upstream_server_t      *server;
	ring                            *hash_ring;
	consistent_hash_conf            *chcf;

	chcf = ngx_http_conf_upstream_srv_conf(
			us,ngx_http_upstream_consistent_hash_module);

	us->peer.init = ngx_http_upstream_init_consistent_hash_peer;

	if (!us->servers) {
		return NGX_ERROR;
	}

	server = us->servers->elts;
	for (n = 0, i = 0; i < us->servers->nelts; i++) {
		n += server[i].naddrs;
	}

	hash_ring = ring_new(chcf->n_replicas,chcf->alias,n,cf,us);

	if (hash_ring == NULL) {
		return NGX_ERROR;
	}

	us->peer.data = hash_ring;

	return NGX_OK;
}

static ngx_int_t ngx_http_upstream_init_consistent_hash_peer(
	ngx_http_request_t* r,
	ngx_http_upstream_srv_conf_t* us) {

	request_ring *data;
	data = request_ring_new((ring*)us->peer.data,r);
	if (data == NULL) {
		return NGX_ERROR;
	}
	r->upstream->peer.free = ngx_http_upstream_free_consistent_hash_peer;
	r->upstream->peer.get = ngx_http_upstream_get_consistent_hash_peer;
	r->upstream->peer.data = data;
	r->upstream->peer.tries = data->ring->n_replicas;
	return NGX_OK;
}

static ngx_int_t ngx_http_upstream_get_consistent_hash_peer(
	ngx_peer_connection_t* pc,
	void* data) {

	request_ring *rr = data;
	rr->peer_connection = pc;

	int n = request_ring_assign(rr);
	if(-1 == n) {
		return NGX_ERROR;
	}

	node *nd = &(rr->ring->nodes[n]);
	ngx_log_debug4(
		NGX_LOG_DEBUG_HTTP,
		pc->log,
		0,
	    "[OLIVER] ID : %s, ASSIGNEE: %s, BALANCED: %ui, HEADER: %s",
	    rr->key.data,
	    nd->name.data,
	    rr->balanced_elsewhere,
	    rr->balanced_header.data);

	pc->cached = 0;
	pc->connection = NULL;
	pc->sockaddr = nd->sockaddr;
	pc->socklen = nd->socklen;
	pc->name = &nd->name;

	return NGX_OK;
}

static void ngx_http_upstream_free_consistent_hash_peer(
		ngx_peer_connection_t *pc,
		void *data,
		ngx_uint_t state) {
	request_ring *rr = data;
	pc->tries = rr->ring->n_replicas - rr->tries;
	return;
}
