from threading import Thread
import os
import mmap
from time import time,sleep
import re
import hashlib
import contextlib

from redis import Redis

from journal import Insertion,Deletion,BaseInsertion,IterStruct
from redis_helper import RedisCount,RedisHash,RedisFIFOQueue,Lock,LockException
from file_like import ReadonlyFile,FileWithClose

class TryAgainSoon(Exception): pass
class Compacting(Exception): pass

DEFAULT_CONTENT_TYPE = 'application/octet-stream'

class Oliver(Thread):
    
    LOOSE_FILE = re.compile(r'^(?P<timestamp>\d+)\.dat$')
    SEPERATOR = '_'
    DATA_FILE_NAME = 'data.dat'
    OFFSET_FILE_NAME = 'offset.dat'
    DELETE_FILE_NAME = 'delete.dat'
    
    
    def __init__(self,
                 path,
                 redis = None,
                 initial_size_bytes = 1000 * 1024,
                 buf_size = 64 * 1024,
                 group_by = lambda key : key):
        
        super(Oliver,self).__init__()
        self.initial_size_bytes = initial_size_bytes
        self.buf_size = buf_size
        self.group_by = group_by
        
        if not os.path.isabs(path):
            raise ValueError('path must be absolute. It was %s' % path)
        
        try:
            os.makedirs(path)
        except OSError:
            pass
        
        # this trick ensures that path ends with a slash
        self._path = os.path.join(path,'')
        # This glob will get all files that are currently being written
        self._writing_glob = '{path}{sep}*'.format(\
                                path = self._path,sep = self.SEPERATOR)
        self._redis = redis or Redis()
        
        self._data_fn = os.path.join(path,self.DATA_FILE_NAME)
        self._offset_fn = os.path.join(path,self.OFFSET_FILE_NAME)
        self._delete_fn = os.path.join(path,self.DELETE_FILE_NAME)
        self._data = None
        
        self.offsets = RedisHash(\
                            self._redis,self._redis_key('OFFSETHASH'),Insertion)
        self._instance_count = RedisCount(\
                            self._redis,self._data_fn)
        self.size = RedisCount(\
                            self._redis,self._redis_key('SIZEHASH'))
        self._delete_queue = RedisFIFOQueue( \
                            self._redis,self._redis_key('DELETEQUEUE'),Deletion)
        self._running = False
        self._put_queue = None
        self._init()
    
    @property
    def _ids(self):
        '''
        Get all the keys in this database
        '''
        return self.offsets.keys()
    
    def __len__(self):
        '''
        Get the number of keys in this database
        '''
        return len(self.offsets)
    
    def head(self,key): 
        '''
        Get metadata about an object, without fetching the object itself
        '''
        return self._get_key(key)
    
    def __delitem__(self,key):
        '''
        Delete an item from the store
        '''
        key = key.lower()
        # TODO: This should be sent to redis as an atomic operation
        del self.offsets[key]
        d = Deletion(_id = key, timestamp = self._timestamp())
        self._delete_queue.push(d)
    
    def __setitem__(self,key,value):
        '''
        Set an item whose value fits easily into memory
        '''
        with FileWithClose(\
                key.lower(),
                DEFAULT_CONTENT_TYPE,
                self._timestamp(),
                self._path,
                self._on_file_close) as f:
            
            f.write(value)
    
    def write_stream(self,key,content_type = DEFAULT_CONTENT_TYPE):
        '''
        Return a write-only file-like object, allowing data to be streamed to a
        key
        '''
        return FileWithClose(\
            key.lower(),
            content_type,
            self._timestamp(),
            self._path,
            self._on_file_close)
    
    def __getitem__(self,key):
        '''
        Get an item whose value is expected to fit easily into memory
        '''
        insert = self._get_key(key)
        offset = insert.offset
        size = insert.size
        end = offset + size
        if end > self._map_size:
            self._remap()
        return self._map[offset : end]
    
    def read_stream(self,key):
        '''
        Return a file-like object for a key, from which data can be streamed
        '''
        insert = self._get_key(key)
        try:
            return ReadonlyFile(self._map,insert)
        except BufferError:
            self._remap()
            return ReadonlyFile(self._map,insert)
    
    def _compaction_proxy(self,*args,**kwargs):
        raise Compacting()
    
    def compact(self):
        '''
        Write to a new file, excluding deleted objects, and older versions of
        existing objects
        '''
        # ensure that this is the only running instance
        if self._instance_count() > 1:
            raise Exception(\
                    'compact may only be called on the sole running instance')
        
        # KLUDGE: This is kinda hacky, but it ensures that no other instances
        # can come online during this operation, since they'll see themselves
        # as the only running instance, try to acquire the write lock,
        # and fail.
        
        # TODO: This isn't atomic.  Write lock should include an option to send
        # other instructions along with the setnx command
        self._instance_count -= 1
        
        with self.write_lock():
            # Disable PUTs and DELETEs
            orig_set = self.__setitem__
            orig_write_stream = self.write_stream
            orig_del = self.__delitem__
            self.__setitem__ = self._compaction_proxy
            self.write_stream = self._compaction_proxy
            self.__delitem__ = self._compaction_proxy
            
            
            offsets = self.offsets.items()
            offsets.sort(key = lambda t : self.group_by(t[0]))
            
            newsize = 0
            new_fn = os.path.join(self._path,'newdata.dat')
            new_ofn = os.path.join(self._path,'newoffsets.dat')
            with contextlib.nested(open(new_fn,'wb'),open(new_ofn,'wb')) as (nf,nof):
                for _id,offset in offsets:
                    rs = self.read_stream(_id)
                    data = rs.read(self.buf_size)
                    size = 0
                    while data:
                        nf.write(data)
                        size += len(data)
                        data = rs.read(self.buf_size)
                    
                    insert = Insertion(\
                            _id = _id,
                            content_type = offset.content_type,
                            offset = newsize,
                            size = size,
                            timestamp = offset.timestamp,
                            digest = offset.digest)
                    nof.write(insert.bytes)
                    newsize += size
                nf.flush()
            
            # Disable GETs and HEADs
            orig_head = self.head
            orig_get = self.__getitem__
            orig_read_stream = self.read_stream
            self.head = self._compaction_proxy
            self.__getitem__ = self._compaction_proxy
            self.read_stream = self._compaction_proxy
            
            
            self.offsets.clear()
            self._map.close()
            self._data.close()
            self._offset.close()
            self._delete.close()
            os.remove(self._offset_fn)
            os.remove(self._data_fn)
            os.remove(self._delete_fn)
            os.rename(new_fn,self._data_fn)
            os.rename(new_ofn,self._offset_fn)
            self._init(has_lock = True)
            
            # Re-enable everything
            self.__setitem__ = orig_set
            self.write_stream = orig_write_stream
            self.__delitem__ = orig_del
            self.head = orig_head
            self.__getitem__ = orig_get
            self.read_stream = orig_read_stream
    
    @property
    def logical_size(self):
        return self.size()
    
    @property
    def actual_size(self):
        return os.stat(self._data_fn).st_size
    
    @property
    def loose_files(self):
        '''
        Return a list of files that aren't currently being written that should
        be moved into the bigfile
        '''
        return sorted(filter(\
                    lambda x : self.LOOSE_FILE.match(x),os.listdir(self._path)))
    
    @property
    def loose_keys(self):
        '''
        Return a set of keys that are currently waiting to be written into the
        bigfile
        '''
        keys = []
        lf = self.loose_files
        for fn in lf:
            with open(os.path.join(self._path,fn),'rb') as f:
                keys.append(BaseInsertion.unpack(f)._id)
        return keys
    
    def _init(self,has_lock = False):
        if has_lock:
            self._create_files()
            self._open_files()
            self._get_offsets_and_deletes()
            self._instance_count += 1
            return
        
        if self._instance_count():
            self._open_files()
            self._instance_count += 1
            return
        
        with self.write_lock():
            self._create_files()
            self._open_files()
            self._get_offsets_and_deletes()
            self._instance_count += 1
    
    def _grow(self,f):
        f.seek(0,2)
        s = 0
        while s < self.initial_size_bytes:
            f.write(b'\x00' * self.buf_size)
            s += self.buf_size
        f.flush()
    
    def _create_files(self):
        if not os.path.exists(self._data_fn) \
            or os.stat(self._data_fn).st_size < self.initial_size_bytes:
            with open(self._data_fn,'a+b') as f:
                self._grow(f)
            
        if not os.path.exists(self._offset_fn):
            with open(self._offset_fn,'wb') as f:
                pass
    
        if not os.path.exists(self._delete_fn):
            with open(self._delete_fn,'wb') as f:
                pass
    
    def _open_files(self):
        self._offset = open(self._offset_fn,'r+b')
        self._delete = open(self._delete_fn,'r+b')
        self._data = open(self._data_fn,'rb')
        self._remap()
    
    def _get_offsets_and_deletes(self):
        self._flush_keys()
        self._get_offsets()
        self._get_deletes()
    
    def _get_offsets(self):
        insertions = IterStruct(self._offset,Insertion)
        size = 0
        for insert in insertions:
            size = insert.offset + insert.size
            self.offsets[insert._id] = insert
        self.size(size)
    
    def _get_deletes(self):
        deletions = IterStruct(self._delete,Deletion)
        for deletion in deletions:
            insertion = self.head(deletion._id)
            if deletion.timestamp > insertion.timestamp:
                del self.offsets[deletion._id]
    
    def _get_key(self,key):
        key = key.lower()
        try:
            return self.offsets[key]
        except KeyError:
            pass
        
        if key in self.loose_keys:
            raise TryAgainSoon(key)
        
        raise KeyError(key)
    
    def key_size(self,key):
        return self._get_key(key).size
    
    def _remap(self):
        self._map = mmap.mmap(self._data.fileno(),0,prot = mmap.PROT_READ)
        self._map_size = self._map.size()
    
    @classmethod
    def _timestamp(cls):
        '''
        Microseconds since the epoch
        '''
        return int(time() * 1e6)
    
    def write_lock(self):
        return Lock(self._redis_key('WRITE'),self._redis)
    
    def _redis_key(self,key):
        return '{key}:{x}'.format(key = key,x = self._offset_fn)
    
    def _flush_keys(self):
        pattern = '*:{x}'.format(x = self._offset_fn)
        keys = self._redis.keys(pattern)
        self._redis.delete(keys)
    
    def _on_file_close(self,f):
        fn = '{timestamp}.dat'.format(**f.__dict__)
        os.rename(f.fp,os.path.join(f.path,fn))
    
    def _parse_filename(self,path):
        _,fn = os.path.split(path)
        timestamp,_ = os.path.splitext(fn)
        return fn,int(timestamp)

    def start(self):
        self._running = True
        Thread.start(self)
    
    def stop(self):
        self._running = False

    def run(self):
        try:
            while self._running:
                sleep(1)
                try:
                    with self.write_lock():
                        self._synch()
                except LockException:
                    pass
        except KeyboardInterrupt:
            pass
    
    def _synch(self):
        self._write_insertions()
        self._write_deletes()
    
    def _write_deletes(self):
        d = self._delete_queue.pop()
        while d:
            self._delete.write(d.bytes)
            self._delete.flush()
            d = self._delete_queue.pop()
    
    def set_put_queue(self,q):
        self._put_queue = q
    
    def _stream_into(self,datafile,fn):
        size = 0
        offset = self.logical_size
        with open(os.path.join(self._path,fn),'rb') as f:
            key = BaseInsertion.unpack(f)
            digest = hashlib.sha1()
            # seek to the logical end of the file
            datafile.seek(self.logical_size)
            data = f.read(self.buf_size)
            ld = len(data)
            while data:
                while self.logical_size + ld > self.actual_size:
                    self._grow(datafile)
                    datafile.seek(self.logical_size)
                datafile.write(data)
                digest.update(data)
                size += ld
                self.size(self.size() + ld)
                data = f.read(self.buf_size)
                ld = len(data)
        
        datafile.flush()
        insertion = Insertion(\
                        _id = key._id,
                        content_type = key.content_type,
                        offset = offset,
                        size = size,
                        timestamp = self._timestamp(),
                        digest = digest.digest())
        
        self._offset.write(insertion.bytes)
        self._offset.flush()
        self.offsets[insertion._id] = insertion
        
        if self._put_queue:
            self._put_queue.add(insertion._id)
    
    def _write_insertions(self):
        # TODO: Re-opening this file each time synch is called is 
        # inefficient.  there will only be one writer at a time, so just
        # keep it open.
        with open(self._data_fn,'r+b') as df:
            lf = [self._parse_filename(f) for f in self.loose_files]
            # sort by ascending timestamp
            lf = sorted(lf,key = lambda x : x[1])
            for fn,timestamp in lf:
                self._stream_into(df,fn)
                os.remove(os.path.join(self._path,fn))
    
    def __enter__(self):
        return self
    
    def __exit__(self,t,value,traceback):
        self.close()
    
    def close(self):
        self.stop()
        self._delete.flush()
        self._map.flush()
        self._offset.flush()
        self._data.close()
        self._offset.close()
        self._delete.close()
        self._instance_count -= 1
