from time import sleep,time
import shutil
from uuid import uuid4
import unittest2
import httplib
import hashlib
from functools import wraps,partial
from random import choice

import requests
import redis

from nginx_conf import *

class RetryAssert(object):
    def __init__(self,assertion,interval = 0.25,timeout = 10):
        self.assertion = assertion
        self.interval = interval
        self.timeout = timeout
    
    def do_assert(self):
        start = time()
        while time() - start < self.timeout:
            try:
                return self.assertion()
            except:
                sleep(self.interval)
        return self.assertion()


secret_key = 'c4797ce2d73b4702b48ec644e91c9e4b'
ip = '127.0.0.1'

class BaseTest(unittest2.TestCase):
    
    GET    = 'GET'
    PUT    = 'PUT'
    DELETE = 'DELETE'
    POST   = 'POST'
    HEAD   = 'HEAD'
    
    JSON = 'application/json'
    
    _auth_header = \
            {'X-Oliver' : hashlib.md5(secret_key + ip).hexdigest() }
    
    def tearDown(self):
        r = redis.Redis()
        r.flushall()
        
        if not self.nodes:
            return
        
        for node in self.nodes:
            node.stop()
            try:
                shutil.rmtree(node.dbpath)
            except OSError:
                pass
    
    def req_assert(\
        self,
        method,
        uri,
        data = None,
        headers = None,
        resp_code = None,
        content = None,
        json_assert = None,
        headers_assert = None,
        content_type = None,
        location = None,
        pass_secret = True):
        
        headers = headers or {}
        if pass_secret:
            headers.update(self._auth_header)
        
        resp = requests.request(method,uri,data = data,headers = headers)
        print "RESPONSE",resp.content
        
        if resp_code:
            self.assertEqual(resp_code,resp.status_code)
        if content:
            self.assertEqual(content,resp.content)
        if json_assert:
            self.assertTrue(json_assert(resp.json()))
        if headers_assert:
            self.assertTrue(headers_assert(resp.headers))
        if content_type:
            self.assertEqual(content_type,resp.headers['Content-Type'])
        if location:
            self.assertEqual(location,resp.headers['Location'])
        return resp
    
    def put_assert(self,uri,content_type,content):
        def x():
            self.req_assert(\
                self.PUT,
                uri,
                data = content,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED,
                location = uri)
        return x
    
    def get_assert(self,uri,content_type,content):
        def x():
            self.req_assert(\
                self.GET,
                uri,
                resp_code = httplib.OK,
                content_type = content_type,
                content = content)
        return x
    
    def del_assert(self,uri):
        def x():
            self.req_assert(\
                self.DELETE,
                uri,
                resp_code = httplib.NO_CONTENT)
        return x
    
    def not_found(self,uri):
        def x():
            self.req_assert(\
                self.GET,
                uri,
                resp_code = httplib.NOT_FOUND)
        return x
    
    def data_uri(self,base,key):
        return data_uri(base,key)

def async(f):
    @wraps(f)
    def x(*args,**kwargs):
        results = []
        funcs = f(*args,**kwargs)
        for func in funcs:
            results.append(RetryAssert(func).do_assert())
        return results
    
    return x

def async_case(cls):
    for k,v in cls.__dict__.iteritems():
        if k.startswith('test_') and callable(v):
            setattr(cls,k,async(v))
    return cls

def data_uri(base,key):
    return '{base}/data/{key}/'.format(**locals())

class Node(object):
    def __init__(self,test_path,**kwargs):
        self.kwargs = kwargs
        self.kwargs['dbpath'] = os.path.join(test_path,uuid4().hex)
        self.kwargs['initialsize'] = str(1000)
        self.kwargs['secret_key'] = secret_key 
        self.kwargs['ip'] = ip
        self.__dict__.update(self.kwargs)
        self.args = ['python','oliver/node.py']
        for k,v in self.kwargs.iteritems():
            self.args.extend(['--' + k, v])
    
    @property
    def base_uri(self):
        return 'http://{host}:{port}'.format(**self.__dict__)
    
    @property
    def address(self):
        return '{host}:{port}'.format(**self.__dict__)
    
    def data_uri(self,key):
        return data_uri(self.base_uri,key)
    
    def start(self):
        self.process = subprocess.Popen(self.args)
    
    def stop(self):
        self.process.terminate()
    
    def __eq__(self,other):
        return self.nodename == other.nodename

@async_case
class NginxOneNode(BaseTest):
    
    def new_node(self,**kwargs):
        node = Node(\
                self.test_path,
                host = 'localhost',
                replicas = '2',
                **kwargs)
        node.start()
        def x():
            self.req_assert(\
                self.GET,
                node.base_uri,
                resp_code = httplib.OK)
        RetryAssert(x).do_assert()
        return node
    
    def setUp(self):
        self.test_path = '/tmp/'
        
        tu1 = TornadoUpstream('frontends1', UpstreamServer('127.0.0.1',9000))
        
        mu = MainUpstream(\
                'main',
                [UpstreamServer('127.0.0.1',8000)],
                2) 
        servers = [ServerConfig(8000,tu1),
                   ServerConfig(80,mu)]
        nginx_config_path = os.path.join(self.test_path,uuid4().hex + '.conf')
        self.config = NginxConfig([tu1],mu,servers,path = nginx_config_path)
        self.config.start()
        
        self.nodea = self.new_node(\
            port = '9000',
            nodename = '127.0.0.1:8000',
            location = 'localhost')
        self.nodes = [self.nodea]
    
    def tearDown(self):
        super(NginxOneNode,self).tearDown()
        self.config.stop()
    
    def test_put(self):
        uri = 'http://localhost/data/ollie/'
        data = 'cute'
        content_type = 'text/plain'
        return [
            self.put_assert(uri,content_type,data),
            self.get_assert(uri,content_type,data)]
    

@async_case
class NginxThreeNodesTwoReplicas(BaseTest):
    
    def new_node(self,**kwargs):
        node = Node(\
                self.test_path,
                host = 'localhost',
                replicas = '2',
                **kwargs)
        node.start()
        def x():
            self.req_assert(\
                self.GET,
                node.base_uri,
                resp_code = httplib.OK)
        RetryAssert(x).do_assert()
        return node
    
    def setUp(self):
        self.test_path = '/tmp/'
        
        tu1 = TornadoUpstream('frontends1', UpstreamServer('127.0.0.1',9000))
        tu2 = TornadoUpstream('frontends2', UpstreamServer('127.0.0.1',9001))
        tu3 = TornadoUpstream('frontends3', UpstreamServer('127.0.0.1',9002))
        
        mu = MainUpstream(\
                'main',
                [UpstreamServer('127.0.0.1',8000),
                 UpstreamServer('127.0.0.1',8001),
                 UpstreamServer('127.0.0.1',8002)],
                2) 
        servers = [ServerConfig(8000,tu1),
                   ServerConfig(8001,tu2),
                   ServerConfig(8002,tu3),
                   ServerConfig(80,mu)]
        nginx_config_path = os.path.join(self.test_path,uuid4().hex + '.conf')
        self.config = NginxConfig([tu1,tu2,tu3],mu,servers,path = nginx_config_path)
        self.config.start()
        
        self.nodea = self.new_node(\
            port = '9000',
            nodename = '127.0.0.1:8000')
        self.nodeb = self.new_node(\
            port = '9001',
            entree = '127.0.0.1:8000',
            nodename = '127.0.0.1:8001')
        self.nodec = self.new_node(\
            port = '9002',
            entree = '127.0.0.1:8000',
            nodename = '127.0.0.1:8002')
        self.nodes = [self.nodea,self.nodeb,self.nodec]
    
    def tearDown(self):
        super(NginxThreeNodesTwoReplicas,self).tearDown()
        self.config.stop()
    
    def test_two_nodes_have_replicas(self):
        key = 'ollie'
        data = 'cute'
        content_type = 'text/plain'
        uri = 'http://localhost/data/' + key
        
        def a():
            self.req_assert(\
                self.PUT,
                uri,
                data = data,
                headers = { 'Content-Type' : content_type },
                resp_code = httplib.CREATED)
        
        def b():
            count = 0
            count += int(requests.get(\
                self.nodea.base_uri + '/data/' + key).status_code == httplib.OK)
            count += int(requests.get(\
                self.nodeb.base_uri + '/data/' + key).status_code == httplib.OK)
            count += int(requests.get(\
                self.nodec.base_uri + '/data/' + key).status_code == httplib.OK)
            self.assertEqual(2,count)
        
        def c():
            for _ in xrange(10):
                self.req_assert(\
                    self.GET,
                    'http://localhost/data/' + key,
                    resp_code = httplib.OK)
        
        return [a,b,c]

@async_case
class NginxTwoNodesTwoReplicas(BaseTest):
    
    def new_node(self,**kwargs):
        node = Node(\
                self.test_path,
                host = 'localhost',
                replicas = '2',
                location = 'localhost',
                **kwargs)
        node.start()
        def x():
            self.req_assert(\
                self.GET,
                node.base_uri,
                resp_code = httplib.OK)
        RetryAssert(x).do_assert()
        return node
    
    def setUp(self):
        self.test_path = '/tmp/'
        
        tu1 = TornadoUpstream('frontends1', UpstreamServer('127.0.0.1',9000))
        tu2 = TornadoUpstream('frontends2', UpstreamServer('127.0.0.1',9001))
        mu = MainUpstream(\
                'main',
                [UpstreamServer('127.0.0.1',8000),
                 UpstreamServer('127.0.0.1',8001)],
                2) 
        servers = [ServerConfig(8000,tu1),
                   ServerConfig(8001,tu2),
                   ServerConfig(80,mu)]
        nginx_config_path = os.path.join(self.test_path,uuid4().hex + '.conf')
        self.config = NginxConfig([tu1,tu2],mu,servers,path = nginx_config_path)
        self.config.start()
        
        self.nodea = self.new_node(\
            port = '9000',
            nodename = '127.0.0.1:8000')
        self.nodeb = self.new_node(\
            port = '9001',
            entree = '127.0.0.1:8000',
            nodename = '127.0.0.1:8001')
        self.nodes = [self.nodea,self.nodeb]
    
    def tearDown(self):
        super(NginxTwoNodesTwoReplicas,self).tearDown()
        self.config.stop()
    
    def test_down_node(self):
        dr = os.path.dirname(os.path.abspath(__file__))
        fn = os.path.join(dr,'down.json')
        with open(fn,'r') as f:
            down_json = f.read()
        
        def a():
            self.nodea.stop()
            self.req_assert(\
                self.GET,
                'http://localhost:8000/data/blah',
                resp_code = httplib.SERVICE_UNAVAILABLE,
                content = down_json)
        return [a]
    
    def test_can_remove_node_and_still_get_data(self):
        key = 'putdata'
        value = 'value'
        content_type = 'text/plain'
        
        def a():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key,
                data = value,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        
        b = self.get_assert(\
                'http://localhost:8000/data/' + key,content_type,value)
        
        b2 = self.get_assert(\
                'http://localhost:8001/data/' + key,content_type,value)
        
        def c():
            self.nodea.stop()
            self.req_assert(\
                self.GET,
                'http://localhost:8000/data/' + key,
                resp_code = httplib.SERVICE_UNAVAILABLE)
        
        def d():
            for _ in xrange(10):
                self.req_assert(\
                    self.GET,
                    'http://localhost/data/' + key,
                    resp_code = httplib.OK,
                    content_type = content_type,
                    content = value)
        
        return [a,b,b2,c,d]
    
    def test_404(self):
        a = self.not_found('http://localhost/data/ollie')
        return [a]
    
    def test_no_keys(self):
        def a():
            resp = self.req_assert(\
                self.GET,
                'http://localhost/data/',
                resp_code = httplib.OK,
                content_type = self.JSON,
                json_assert = lambda j : not j['keys'])
        
        return [a]
    
    def test_can_put_data(self):
        key = 'putdata'
        value = 'value'
        content_type = 'text/plain'
        def a():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key,
                data = value,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        
        b = self.get_assert(\
                'http://localhost/data/' + key,content_type,value)
        
        return [a,b]
    
    def test_can_put_data_using_chunked_transfer_encoding(self):
        key = 'chunked'
        value = (x for x in ['val','ue'])
        content_type = 'text/plain'
        def a():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key,
                data = value,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        
        b = self.get_assert(\
                'http://localhost/data/' + key,content_type,'value')
        
        return [a,b]
    
    def test_can_get_partial_data(self):
        key = 'partial'
        value = 'value'
        content_type = 'text/plain'
        def a():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key,
                data = value,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        def b():
            self.req_assert(\
                self.GET,
                'http://localhost/data/' + key,
                headers = {'Range' : 'bytes=1-3'},
                resp_code = httplib.PARTIAL_CONTENT,
                content_type = content_type,
                content = 'alu')
        
        return [a,b]
    
    def test_can_delete_data(self):
        key = 'ollie'
        value = 'value'
        content_type = 'text/plain'
        uri = self.data_uri('http://localhost', key)
        a = self.put_assert(uri,content_type,value)
        b = self.get_assert(uri,content_type,value)
        c = self.del_assert(uri)
        d = self.not_found(uri)
        
        return [a,b,c,d]
    
    def test_can_update_data(self):
        key = 'update'
        value = 'value'
        content_type = 'text/plain'
        def a():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key,
                data = value,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        
        b = self.get_assert(\
                'http://localhost/data/' + key,content_type,value)
            
        def c():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key,
                data = value + 'CHANGED',
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        
        def d():
            self.req_assert(\
                self.GET,
                'http://localhost/data/' + key,
                resp_code = httplib.OK,
                content_type = content_type,
                content = value + 'CHANGED')
        
        return [a,b,c,d]
    
    
    def test_can_get_304_for_unchanged_data(self):
        key = 'unchanged'
        value = 'value'
        content_type = 'text/plain'
        def a():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key,
                data = value,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        
        def b():
            resp = self.req_assert(\
                self.GET,
                'http://localhost/data/' + key,
                resp_code = httplib.OK,
                content_type = content_type,
                content = value)
            etag = resp.headers['etag']
            self.req_assert(\
                self.GET,
                'http://localhost/data/' + key,
                headers = {'If-None-Match' : etag},
                resp_code = httplib.NOT_MODIFIED)
        
        return [a,b]
    
    def test_can_get_keys(self):
        key = 'getkeys'
        value = 'value'
        content_type = 'text/plain'
        
        def a():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key,
                data = value,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        
        key2 = 'key2'
        value2 = 'value2'
        def b():
            self.req_assert(\
                self.PUT,
                'http://localhost/data/' + key2,
                data = value2,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED)
        
        def c():
            self.req_assert(\
                self.GET,
                'http://localhost/data/',
                resp_code = httplib.OK,
                content_type = self.JSON,
                json_assert = lambda j : set([key,key2]) == set(j['keys']))
        
        return [a,b,c]

@async_case
class ThreeNodesTwoReplicas(BaseTest):
    
    port = 9333
    
    def new_node(self,**kwargs):
        node = Node(\
                self.test_path,
                host = 'localhost',
                port = str(self.port),
                replicas = '2',
                **kwargs)
        self.port += 1
        node.start()
        def x():
            self.req_assert(\
                self.GET,
                node.base_uri,
                resp_code = httplib.OK)
        RetryAssert(x).do_assert()
        return node
    
    def setUp(self):
        self.test_path = '/tmp/'
        self.nodes = []
        nodea = self.new_node(nodename = 'nodea')
        self.nodes.append(nodea)
        self.nodes.append(\
                    self.new_node(nodename = 'nodeb', entree = nodea.address))
        self.nodes.append(\
                    self.new_node(nodename = 'nodec', entree = nodea.address))
    
    def test_data_is_replicated(self):
        node = choice(self.nodes)
        key = 'Ollie'
        data = 'blah'
        uri = node.base_uri + '/data/' + key
        
        def a():
            self.req_assert(\
                self.PUT,
                uri,
                data = data,
                headers = {'Content-Type' : 'text/plain'},
                resp_code = httplib.CREATED)
            
        def b():
            count = 0
            for node in self.nodes:
                resp = requests.get(node.base_uri + '/data/' + key)
                ok = httplib.OK == resp.status_code
                val = data == resp.content
                count += int(ok and val)
            self.assertEqual(2,count)
        
        return [a,b]
    
    def test_deletes_propagate(self):
        node = choice(self.nodes)
        key = 'Ollie'
        data = 'blah'
        uri = node.base_uri + '/data/' + key
        
        def a():
            self.req_assert(\
                self.PUT,
                uri,
                data = data,
                headers = {'Content-Type' : 'text/plain'},
                resp_code = httplib.CREATED)
        
        def b():
            count = 0
            for node in self.nodes:
                resp = requests.get(node.base_uri + '/data/' + key)
                ok = httplib.OK == resp.status_code
                val = data == resp.content
                count += int(ok and val)
            self.assertEqual(2,count)
        
        c = self.del_assert(uri)
        
        def d():
            count = 0
            for node in self.nodes:
                resp = requests.get(node.base_uri + '/data/' + key)
                not_found = httplib.NOT_FOUND == resp.status_code
                count += int(not_found)
            self.assertEqual(3,count)
        
        return [a,b,c,d]
    
    def test_node_down_propagates(self):
        def a():
            for node in self.nodes:
                self.req_assert(\
                    self.GET,
                    node.base_uri,
                    resp_code = httplib.OK,
                    content_type = self.JSON,
                    json_assert = lambda j : 2 == len(j['nodes']))
        
        def b():
            self._down = choice(self.nodes)
            self._down.stop()    
            
        def c():
            for node in filter(lambda x : x != self._down,self.nodes):
                self.req_assert(\
                    self.GET,
                    node.base_uri,
                    resp_code = httplib.OK,
                    content_type = self.JSON,
                    json_assert = lambda j : 1 == len(j['nodes']))
        
        return[a,b,c]


@async_case
class TwoNodesTwoReplicas(BaseTest):
    
    port = 9333
    
    def new_node(self,**kwargs):
        node = Node(\
                self.test_path,
                host = 'localhost',
                port = str(self.port),
                replicas = '2',
                **kwargs)
        self.port += 1
        node.start()
        def x():
            self.req_assert(\
                self.GET,
                node.base_uri,
                resp_code = httplib.OK)
        RetryAssert(x).do_assert()
        return node
    
    def setUp(self):
        self.test_path = '/tmp/'
        self.nodes = []
        nodea = self.new_node(nodename = 'nodea')
        self.nodes.append(nodea)
        self.nodes.append(\
                    self.new_node(nodename = 'nodeb', entree = nodea.address))
    
    def test_each_node_knows_about_the_other(self):
        nodea,nodeb = self.nodes
        namea,nameb = \
            [hashlib.md5(node.nodename).hexdigest() for node in self.nodes]
        
        def a():
            resp = requests.get(nodea.base_uri, headers = self._auth_header)
            self.assertEqual(httplib.OK,resp.status_code)
            self.assertEqual('application/json',resp.headers['Content-Type'])
            node_dict = resp.json()['nodes']
            self.assertEqual(1,len(node_dict))
            self.assertTrue(nameb in node_dict)
            self.assertEqual([nodeb.host,int(nodeb.port)],node_dict[nameb])
        
        def b():
            resp = requests.get(nodeb.base_uri,headers = self._auth_header)
            self.assertEqual(httplib.OK,resp.status_code)
            self.assertEqual('application/json',resp.headers['Content-Type'])
            node_dict = resp.json()['nodes']
            self.assertEqual(1,len(node_dict))
            self.assertTrue(namea in node_dict)
            self.assertEqual([nodea.host,int(nodea.port)],node_dict[namea])
        
        return [a,b]
    
    def test_data_is_replicated(self):
        nodea,nodeb = self.nodes
        key = 'ollie'
        data = 'data'
        nodea_uri = self.data_uri(nodea.base_uri, key)
        nodeb_uri = self.data_uri(nodeb.base_uri, key)
        content_type = 'text/plain'
        a = self.put_assert(nodea_uri, content_type, data)
        b = self.get_assert(nodea_uri,content_type,data)
        c = self.get_assert(nodeb_uri,content_type,data)
        return [a,b,c]

    def test_data_is_replicated_with_correct_content_type(self):
        nodea,nodeb = self.nodes
        key = 'ollie'
        data = 'data'
        content_type = 'text/plain'
        nodea_uri = self.data_uri(nodea.base_uri, key)
        nodeb_uri = self.data_uri(nodeb.base_uri, key)
        a = self.put_assert(nodea_uri,content_type,data)
        b = self.get_assert(nodeb_uri,content_type,data)
        return [a,b]
    
    def test_deletes_propagate(self):
        nodea,nodeb = self.nodes
        key = 'ollie'
        data = 'data'
        
        def a():
            self.req_assert(\
                self.PUT,
                nodea.base_uri + '/data/' + key,
                data = data,
                headers = {'Content-Type' : 'text/plain'},
                resp_code = httplib.CREATED)
        
        b = self.get_assert(\
                nodea.base_uri + '/data/' + key,'text/plain',data)
        c = self.get_assert(\
                nodeb.base_uri + '/data/' + key,'text/plain',data)
        d = self.del_assert(nodea.base_uri + '/data/' + key)
        e = self.not_found(nodea.base_uri + '/data/' + key)
        f = self.not_found(nodeb.base_uri + '/data/' + key)
        
        return [a,b,c,d,e,f]
            

@async_case
class SingleNode(BaseTest):
    
    port = 9333
    
    def new_node(self,nodename = 'oliver',public_addr = ''):
        self.node = Node(\
                    self.test_path,
                    host = 'localhost',
                    port = str(self.port),
                    nodename = nodename,
                    public_addr = public_addr)
        self.port += 1
        self.node.start()
    
    def setUp(self):
        self.test_path = '/tmp/'
        self.new_node()
        
    def tearDown(self):
        if not self.node:
            return
        self.node.stop()
        try:
            shutil.rmtree(self.node.dbpath)
        except OSError:
            pass
        r = redis.Redis()
        r.flushall()
    
    @property
    def base_uri(self):
        return self.node.base_uri
    
    def put_then_assert(self,uri,data,content_type):
        def a():
            self.req_assert(
                self.PUT,
                uri,
                data,
                {'Content-Type' : content_type},
                resp_code = httplib.CREATED,
                location = uri)
        return a
    
    def test_handshake_host_and_port_are_public_when_specified(self):
        # yes, this is a total hack
        self.node.stop()
        self.new_node(nodename = 'oliver',public_addr = 'zing:9000')
        def a(j):
            return j['host'] == 'zing' and j['port'] == 9000
         
        def x():
            self.req_assert(\
                self.GET,
                self.node.base_uri,
                json_assert = a)
        return [x]
                
    def test_main(self):
        def x():
            self.req_assert(\
                self.GET,
                self.base_uri,
                resp_code = httplib.OK)
        return [x]
    
    def test_main_content_type(self):
        def x():
            self.req_assert(\
                self.GET,
                self.base_uri,
                content_type = self.JSON)
        return [x]
    
    def test_main_data(self):
        nodename = 'oliver'
        nodehash = hashlib.md5(nodename).hexdigest()
        def x():
            self.req_assert(\
                self.GET,
                self.base_uri,
                json_assert = lambda j : \
                    nodehash == j['nodename'] and 1 == j['n_replicas'])
        return [x]
    
    def test_node_get(self):
        def x():
            self.req_assert(\
                self.GET,
                self.base_uri + '/nodes',
                resp_code = httplib.OK,
                content_type = self.JSON,
                json_assert = lambda j : not j)
        return [x]
    
    def test_node_404(self):
        x = self.not_found(self.base_uri + '/nodes/key')
        return [x]
    
    def test_data_get_empty(self):
        def x():
            self.req_assert(\
                self.GET,
                self.base_uri + '/data/',
                resp_code = httplib.OK,
                content_type = self.JSON,
                json_assert = lambda j : not j['keys'])
        return [x]
    
    def test_data_put(self):
        uri = self.base_uri + '/data/ollie/'
        data = 'blah'
        content_type = 'text/plain'
        return [self.put_then_assert(uri, data, content_type)]
    
    def test_data_put_chunked(self):
        uri = self.base_uri + '/data/ollie/'
        chunks = ['this','is','chunked','data']
        content_type = 'text/plain'
        
        def x():
            data = (x for x in chunks)
            self.req_assert(\
                self.PUT,
                uri,
                data = data,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED,
                location = uri)
        return [x]
    
    def test_data_get(self):
        uri = self.base_uri + '/data/ollie/'
        data = u'blah'
        content_type = 'text/plain'
        
        x = self.put_then_assert(uri, data, content_type)
        y = self.get_assert(uri,content_type,data)
        
        return [x,y]
    
    def test_data_get_after_chunked_put(self):
        uri = self.base_uri + '/data/ollie/'
        chunks = ['this','is','chunked','data']
        content_type = 'text/plain'
        
        def x():
            data = (x for x in chunks)
            self.req_assert(\
                self.PUT,
                uri,
                data = data,
                headers = {'Content-Type' : content_type},
                resp_code = httplib.CREATED,
                location = uri)
        
        y = self.get_assert(uri,content_type,string.join(chunks,''))
        
        return [x,y]
    
    def test_keys_get(self):
        data = {'key1' : 'data1', 'key2' : 'data2'}
        content_type = 'text/plain'
        
        funcs = []
        
        for k,v in data.iteritems():
            uri = '{base}/data/{key}/'.format(base = self.base_uri,key = k)
            def x(uri,v): 
                self.req_assert(\
                    self.PUT,
                    uri,
                    data = v,
                    headers = {'Content-Type' : content_type},
                    resp_code = httplib.CREATED,
                    location = uri)
            funcs.append(partial(x,uri,v))
        
        def y():
            self.req_assert(\
                self.GET,
                self.base_uri + '/data/',
                resp_code = httplib.OK,
                content_type = self.JSON,
                json_assert = lambda j : set(data.keys()) == set(j['keys']))
        
        funcs.append(y)
        return funcs
    
    def test_data_get_404(self):
        x = self.not_found(self.base_uri + '/data/blah/')
        return [x]
    
    def test_data_head(self):
        uri = self.base_uri + '/data/ollie/'
        data = u'blah'
        content_type = 'text/plain'
    
        a = self.put_then_assert(uri, data, content_type)
        
        def b():
            self.req_assert(\
                self.HEAD,
                uri,
                resp_code = httplib.NO_CONTENT)
        return [a,b]
    
    def test_data_head_404(self):
        def x():
            self.req_assert(\
                self.HEAD,
                self.base_uri + '/data/blah/',
                resp_code = httplib.NOT_FOUND)
        return [x]
    
    def test_get_byte_range_open(self):
        uri = self.base_uri + '/data/ollie/'
        data = u'blah'
        content_type = 'text/plain'
        
        a = self.put_then_assert(uri, data, content_type)
        
        def b():
            self.req_assert(\
                self.GET,
                uri,
                headers = {'Range' : 'bytes=1-'},
                resp_code = httplib.PARTIAL_CONTENT,
                content = 'lah')
        return [a,b]
            
    
    def test_get_byte_range_closed(self):
        uri = self.base_uri + '/data/ollie/'
        data = u'blah'
        content_type = 'text/plain'
        
        a = self.put_then_assert(uri, data, content_type)
        
        def b():
            self.req_assert(\
                self.GET,
                uri,
                headers = {'Range' : 'bytes=1-2'},
                resp_code = httplib.PARTIAL_CONTENT,
                content = 'la')
        
        return [a,b]
    
    def test_data_delete(self):
        uri = self.base_uri + '/data/ollie/'
        data = 'blah'
        content_type = 'text/plain'
        
        a = self.put_then_assert(uri, data, content_type)
        b = self.get_assert(uri,content_type,data)
        c = self.del_assert(uri)
        d = self.not_found(uri)
        
        return [a,b,c,d]
    
    def test_403_when_putting_data_without_secret(self):
        def x():
            self.req_assert(\
                self.PUT,
                self.base_uri + '/data/ollie',
                data = 'cute',
                headers = {'Content-Type' : 'text/plain'},
                resp_code = httplib.FORBIDDEN,
                pass_secret = False)
        return [x]
    
    def test_403_when_getting_root_without_secret(self):
        def x():
            self.req_assert(\
                self.GET,
                self.base_uri,
                resp_code = httplib.FORBIDDEN,
                pass_secret = False)
        return [x]
    
    def test_403_when_getting_nodes_without_secret(self):
        def x():
            self.req_assert(\
                self.GET,
                self.base_uri + '/nodes/',
                resp_code = httplib.FORBIDDEN,
                pass_secret = False)
        return [x]
    
    def test_403_when_putting_node_without_secret(self):
        def x():
            self.req_assert(\
                self.PUT,
                self.base_uri + '/nodes/blah',
                data = {},
                resp_code = httplib.FORBIDDEN,
                pass_secret = False)
        return [x]
    
    def test_403_when_deleting_node_without_secret(self):
        def x():
            self.req_assert(\
                self.DELETE,
                self.base_uri + '/nodes/blah',
                resp_code = httplib.FORBIDDEN,
                pass_secret = False)
        return [x]