import os.path
from uuid import uuid4
import httplib
import hashlib
from random import randint,shuffle
import re

from test_integration import BaseTest,Node,async_case
from nginx_conf import *
from node import Ring



class RealWorldTest(BaseTest):
    delimiter = re.compile('=+')
    
    '''
    Test cases that are failing when I try to set this up in the real world
    '''
    
    def setUp(self):
        self.nodes = None
    
    def test_two_nodes_one_replica_should_fetch_data_from_either_node(self):
        '''
        I get a 404 when trying to access a key that isn't hosted on the server
        I'm requesting it from.  Any node should be able to serve any data 
        request, regardless of whether the data is local or remote
        '''
        self.fail()
    
    def test_two_nodes_two_replicas_both_nodes_should_have_all_keys(self):
        '''
        Requests for data keys hang.  I'm assuming I've written some kind of
        infinite loop where two instances are just redirecting back and forth
        endlessly
        '''
        self.fail()
    
    
    def test_nodes_are_sorted_the_same_by_python_and_nginx(self):
        self.ip = '127.0.0.1'
        ports = range(8000,9000)
        shuffle(ports)
        ports = ports[:20]
        addrs = ['127.0.0.1:%i' % p for p in ports]
        
        upstreams = [UpstreamServer(*addr.split(':')) for addr in addrs]
        self.config_path = os.path.join('/tmp/',uuid4().hex + '.conf') 
        self.public_addr = self.ip
        mu = MainUpstream('main',upstreams,1) 
        servers = [ServerConfig(80,mu)]
        self.config = NginxConfig(\
                    [],
                    mu,
                    servers,
                    path = self.config_path,
                    alias = self.public_addr)
        
        addr2 = list(addrs)
        addr2[0] = self.ip
        d = dict([(hashlib.md5(addr).hexdigest(),addr) for addr in addr2])
        out = self.delimiter.split(self.config.start())[0]
        out = filter(lambda x : x,out.split('\n'))
        keys = d.keys()
        keys.sort()
        keys = [d[k] for k in keys]
        self.assertEqual(out,keys)
    
    def test_keys_are_assigned_to_the_same_nodes_by_python_and_nginx(self):
        '''
        the nginx load balancer and python Ring class should both assign the same
        key to the same node.
        '''
        self.ip = '127.0.0.1'
        ports = range(8000,9000)
        shuffle(ports)
        ports = ports[:20]
        addrs = ['127.0.0.1:%i' % p for p in ports]
        
        upstreams = [UpstreamServer(*addr.split(':')) for addr in addrs]
        self.config_path = os.path.join('/tmp/',uuid4().hex + '.conf') 
        self.public_addr = self.ip
        mu = MainUpstream('main',upstreams,1) 
        servers = [ServerConfig(80,mu)]
        self.config = NginxConfig(\
                    [],
                    mu,
                    servers,
                    path = self.config_path,
                    alias = self.public_addr)
        
        addr2 = list(addrs)
        addr2[0] = self.ip
        d = dict([(hashlib.md5(addr).hexdigest(),addr) for addr in addr2])
        out,assignments,ids,_ = self.delimiter.split(self.config.start())
        out = filter(lambda x : x,out.split('\n'))
        assignments = [int(i) for i in filter(lambda x : x,assignments.split('\n'))]
        ids = filter(lambda x : x,ids.split('\n'))
        r = Ring(d.keys())
        python_assignments = [r.bisect(hashlib.md5(x).hexdigest()) for x in ids]
        self.assertEqual(python_assignments,assignments)
        
    
    def test_local_node_indices_are_the_same(self):
        self.ip = '127.0.0.1'
        ports = range(8000,9000)
        shuffle(ports)
        ports = ports[:20]
        addrs = ['127.0.0.1:%i' % p for p in ports]
        
        upstreams = [UpstreamServer(*addr.split(':')) for addr in addrs]
        self.config_path = os.path.join('/tmp/',uuid4().hex + '.conf') 
        self.public_addr = self.ip
        mu = MainUpstream('main',upstreams,1) 
        servers = [ServerConfig(80,mu)]
        self.config = NginxConfig(\
                    [],
                    mu,
                    servers,
                    path = self.config_path,
                    alias = self.public_addr)
        
        addr2 = list(addrs)
        addr2[0] = self.ip
        d = dict([(hashlib.md5(addr).hexdigest(),addr) for addr in addr2])
        nginx_local_node_index = int(self.delimiter.split(self.config.start())[3])
        keys = d.keys()
        keys.sort()
        keys = [d[k] for k in keys]
        python_local_node_index = keys.index(self.ip)
        self.assertEqual(python_local_node_index,nginx_local_node_index)
        
    
    

@async_case
class LoadBalancerTest(BaseTest):
    
    def setUp(self):
        upstreams = [UpstreamServer('localhost',8000)]
        self.ip = '127.0.0.1'
        self.config_path = os.path.join('/tmp/',uuid4().hex + '.conf') 
        self.public_addr = self.ip
        mu = MainUpstream('main',upstreams,1) 
        servers = [ServerConfig(80,mu)]
        self.config = NginxConfig(\
                    [],
                    mu,
                    servers,
                    path = self.config_path,
                    alias = self.public_addr)
        self.config.start()
        
        node = Node(\
                '/tmp/',
                host = 'localhost',
                port = '8000',
                replicas = '1',
                public_addr = self.public_addr)
        node.start()
        self.nodes = [node]
    
    def tearDown(self):
        super(LoadBalancerTest,self).tearDown()
        self.config.stop()
    
    def test_x_oliver_balanced_header_should_be_set_when_not_included(self):
        def x():
            self.req_assert(\
                self.GET,
                'http://localhost/debug',
                resp_code = httplib.OK,
                json_assert = lambda j : j['headers']['X-Oliver-Balanced'] == '127.0.0.1')
        return [x]
    
    def test_x_oliver_balanced_header_should_be_appended_to_when_included(self):
        def x():
            self.req_assert(\
                self.GET,
                'http://localhost/debug',
                headers = {'X-Oliver-Balanced' : 'blah'},
                resp_code = httplib.OK,
                json_assert = lambda j : j['headers']['X-Oliver-Balanced'] == '127.0.0.1,blah')
        return [x]
    