import unittest2
import httplib
import os
from uuid import uuid4
import shutil
import hashlib

from node import Node,RangeRequest

class Bag(object):
    def __init__(self,**kwargs):
        self.__dict__.update(kwargs)

class RangeRequestTests(unittest2.TestCase):
    
    def _headers(self,r = None):
        return { 'Range' : r } if r else {}
    
    def test_response_code_is_200_if_no_range(self):
        rr = RangeRequest(self._headers(), Bag(size = 9))
        self.assertEqual(httplib.OK,rr.response_code)
    
    def test_response_code_is_206_if_range(self):
        rr = RangeRequest(self._headers(r = 'bytes=0-'), Bag(size = 9))
        self.assertEqual(httplib.PARTIAL_CONTENT,rr.response_code)
    
    # KLUDGE: Is this the correct response for a malformed byte range request?
    def test_response_code_is_416_if_range_not_parseable(self):
        rr = RangeRequest(\
                    self._headers(r = 'bytes=banana-apple'), Bag(size = 9))
        self.assertEqual(\
                    httplib.REQUESTED_RANGE_NOT_SATISFIABLE,rr.response_code)
    
    def test_response_code_is_416_if_range_cannot_be_satisfied(self):
        rr = RangeRequest(self._headers(r = 'bytes=11-21'), Bag(size = 10))
        self.assertEqual(\
                    httplib.REQUESTED_RANGE_NOT_SATISFIABLE,rr.response_code)
    
    def test_start_equals_specified_index(self):
        rr = RangeRequest(self._headers(r = 'bytes=13-20'), Bag(size = 30))
        self.assertEqual(13,rr.start)
    
    def test_stop_equals_specified_index(self):
        rr = RangeRequest(self._headers(r = 'bytes=13-20'), Bag(size = 30))
        self.assertEqual(20,rr.stop)
    
    def test_stop_equals_end_of_file_for_open_range(self):
        rr = RangeRequest(self._headers(r = 'bytes=13-'), Bag(size = 30))
        self.assertEqual(29,rr.stop)
    
    def test_content_length_equals_entity_size_if_no_range(self):
        rr = RangeRequest(self._headers(), Bag(size = 30))
        self.assertEqual(30,rr.content_length)
    
    def test_content_length_is_correct_for_open_range(self):
        rr = RangeRequest(self._headers(r = 'bytes=13-'), Bag(size = 30))
        self.assertEqual(17,rr.content_length)
    
    def test_content_length_is_correct_for_closed_range(self):
        rr = RangeRequest(self._headers(r = 'bytes=13-20'), Bag(size = 30))
        self.assertEqual(8,rr.content_length)
    
    def test_content_range_is_none_if_no_range(self):
        rr = RangeRequest(self._headers(), Bag(size = 30))
        self.assertEqual(None,rr.content_range)
    
    def test_content_range_is_correct_for_open_range(self):
        rr = RangeRequest(self._headers(r = 'bytes=13-'), Bag(size = 30))
        self.assertEqual('bytes 13-29/30',rr.content_range)
    
    def test_content_range_is_correct_for_closed_range(self):
        rr = RangeRequest(self._headers(r = 'bytes=13-20'), Bag(size = 30))
        self.assertEqual('bytes 13-20/30',rr.content_range)

class NodeTests(unittest2.TestCase):
    
    def tearDown(self):
        try:
            shutil.rmtree(self.path)
        except OSError:
            pass
    
    def build_node(self, nodename = None):
        dirname = uuid4().hex
        self.path = os.path.join('/tmp/',dirname)
        return Node( \
                 'localhost',
                 port         = 8080,
                 dbpath       = self.path,
                 initial_size = 1024,
                 nodename     = nodename,
                 ip = '127.0.0.1',
                 secret_key = uuid4().hex)
    
    def test_node_id_should_be_hashed_host_and_port_when_no_name_is_provided(self):
        with self.build_node(nodename = None) as node:
            self.assertEqual(hashlib.md5('{host}:{port}'\
            .format(host = node.host,port = node.port)).hexdigest(),node._id)
    
    def test_node_id_should_be_hashed_name_when_name_is_provided(self):
        nodename = 'oliver'
        with self.build_node(nodename = nodename) as node:
            self.assertEqual(hashlib.md5(nodename).hexdigest(),node._id)
    
    
    