import unittest2
import shutil
from uuid import uuid4
import os.path
from time import sleep
from redis import Redis
 
from db import Oliver,TryAgainSoon

class BaseOliverTests(object):
    
    @classmethod
    def setUpClass(cls):
        cls.redis = Redis()
    
    def setUp(self):
        self._dir = uuid4().hex
        self._path = os.path.join('/tmp/',self._dir)
        self.clear_keys() 
        
    def tearDown(self):
        try:
            shutil.rmtree(self._path)
        except OSError:
            # the directory wasn't created
            pass
        self.clear_keys()
    
    def clear_keys(self):
        # KLUDGE: This is a hack.  Knowing how Oliver stores keys in redis
        # is an implementation detail that the tests shouldn't be privy to.
        keys = self.redis.keys('*{x}*'.format(x = self._dir))
        if keys:
            self.redis.delete(*keys)
    
    @property
    def redis(self):
        return self.__class__.redis
    
    def instance(self,
                 path = None,
                 initial_size_bytes = None,
                 buf_size = None,
                 group_by = None):
        
        return Oliver(\
                    path = path or self._path,
                    redis = self.redis,
                    initial_size_bytes = initial_size_bytes or 1000 * 1024,
                    buf_size = buf_size or 64 * 1024,
                    group_by = lambda key : key)
    
    def random_key(self):
        return uuid4().hex
    
    def random_bytes(self,length):
        return os.urandom(length)
    
class OliverInterfaceTests(BaseOliverTests,unittest2.TestCase):
    
    def test_count_should_be_zero_after_first_init(self):
        with self.instance() as db:
            self.assertEqual(0,len(db))
    
    def test_keys_should_be_empty_after_first_init(self):
        with self.instance() as db:
            self.assertEqual(0,len(db._ids))
        
    def test_should_raise_if_path_is_relative(self):
        self.assertRaises(\
                    ValueError,lambda : self.instance(path = self._dir))
    
    def test_should_get_key_error_if_read_nonexistent_key(self):
        with self.instance() as db:
            self.assertRaises(KeyError,lambda : db['SomeKey'])
    
    def test_should_get_key_error_if_stream_nonexistent_key(self):
        with self.instance() as db:
            self.assertRaises(KeyError,lambda : db.read_stream('SomeKey'))
        pass
    
    def test_should_get_TryAgainSoon_if_reading_before_synch(self):
        with self.instance() as db:
            k = self.random_key()
            db[k] = self.random_bytes(10)
            self.assertRaises(TryAgainSoon,lambda : db[k])
    
    def test_should_get_TryAgainSoon_if_streaming_before_synch(self):
        with self.instance() as db:
            k = self.random_key()
            db[k] = self.random_bytes(10)
            self.assertRaises(TryAgainSoon,lambda : db.read_stream(k))
    
    def test_can_get_and_set_item_synchronously(self):
        with self.instance() as db:
            k = self.random_key()
            v = self.random_bytes(100)
            db[k] = v
            db._synch()
            self.assertEqual(v,db[k])
    
    def test_can_set_synchronously_and_then_stream(self):
        with self.instance() as db:
            k = self.random_key()
            size = 100
            v = self.random_bytes(size)
            db[k] = v
            db._synch()
            self.assertEqual(v,db.read_stream(k).read(size))
    
    def test_can_stream_then_synchronously_read(self):
        with self.instance() as db:
            k = self.random_key()
            size = (64 * 1024) + 11
            v = self.random_bytes(size)
            with db.write_stream(k) as stream:
                stream.write(v)
            db._synch()
            self.assertEqual(v,db[k])
    
    def test_can_stream_into_and_out_of(self):
        with self.instance() as db:
            k = self.random_key()
            size = (64 * 1024) + 11
            v = self.random_bytes(size)
            with db.write_stream(k) as stream:
                stream.write(v)
            db._synch()
            self.assertEqual(v,db.read_stream(k).read(size))
    
    def test_can_set_then_close_then_reopen_then_read(self):
        k = self.random_key()
        size = 100
        v = self.random_bytes(size)
        
        with self.instance() as db:    
            db[k] = v
            db._synch()
        
        with self.instance() as db:
            self.assertEqual(v,db[k])
    
    def test_count_returns_correct_value(self):
        with self.instance() as db:
            db['Oliver'] = self.random_bytes(100)
            db['Rebekah'] = self.random_bytes(1025)
            db._synch()
            self.assertEqual(2,len(db))
        
    def test_new_value_overwrites_old_value(self):
        with self.instance() as db:
            db['Oliver'] = 'Old'
            db._synch()
            db['Oliver'] = 'New'
            db._synch()
            self.assertEqual('New',db['Oliver'])
    
    def test_count_returns_correct_value_after_overwrite(self):
        with self.instance() as db:
            db['Oliver'] = 'Old'
            db._synch()
            db['Oliver'] = 'New'
            db._synch()
            self.assertEqual(1,len(db))
    
    def test_keys_returns_correct_value_after_overwrite(self):
        with self.instance() as db:
            db['Oliver'] = 'Old'
            db._synch()
            db['Oliver'] = 'New'
            db._synch()
            self.assertEqual(set(['oliver']),set(db._ids))
    
    def test_key_error_after_deletion(self):
        with self.instance() as db:
            db['Oliver'] = 'Old'
            db._synch()
            del db['Oliver']
            self.assertRaises(KeyError, lambda : db['oliver'])
    
    def test_count_returns_correct_value_after_deletion(self):
        with self.instance() as db:
            db['Oliver'] = 'Old'
            db._synch()
            del db['Oliver']
            self.assertEqual(0,len(db))
    
    def test_keys_returns_correct_value_after_deletion(self):
        with self.instance() as db:
            db['Oliver'] = self.random_bytes(100)
            db['Rebekah'] = self.random_bytes(1025)
            db._synch()
            del db['Rebekah']
            self.assertEqual(set(['oliver']),set(db._ids))
    
    def test_compact_sorts_keys(self):
        with self.instance() as db:
            av = self.random_bytes(10)
            bv = self.random_bytes(10)
            cv = self.random_bytes(10)
            dv = self.random_bytes(10)
            db['b'] = bv
            db['a'] = av
            db['c'] = cv
            db['d'] = dv
            db._synch()
            self.assertEqual(0,db.head('b').offset)
            self.assertEqual(10,db.head('a').offset)
            self.assertEqual(20,db.head('c').offset)
            self.assertEqual(30,db.head('d').offset)
            del db['c']
            db.compact()
            self.assertEqual(3,len(db))
            self.assertEqual(0,db.head('a').offset)
            self.assertEqual(10,db.head('b').offset)
            self.assertEqual(20,db.head('d').offset)
            self.assertEqual(av,db['a'])
            self.assertEqual(bv,db['b'])
            self.assertEqual(dv,db['d'])
            
    def test_compact_should_raise_when_there_are_multiple_instances(self):
        db1 = self.instance()
        sleep(1)
        db2 = self.instance()
        self.assertRaises(Exception,lambda : db1.compact())
    
    def test_head_returns_correct_size_information(self):
        with self.instance() as db:
            db['Oliver'] = self.random_bytes(9)
            db._synch()
            self.assertEqual(9,db.head('Oliver').size)
    
    def test_keys_are_case_insensitive(self):
        with self.instance() as db:
            v = self.random_bytes(13)
            db['Oliver'] = v
            db._synch()
            _ids = db._ids
            self.assertTrue('oliver' in _ids)
            self.assertEqual(v,db['OLivEr'])
    
    def test_content_type_should_default_to_octet_stream(self):
        with self.instance() as db:
            db['oliver'] = self.random_bytes(100)
            db._synch()
            meta = db.head('oliver')
            self.assertEqual('application/octet-stream',meta.content_type)
    
    def test_user_specified_content_type_should_be_returned_by_head(self):
        with self.instance() as db:
            with db.write_stream('oliver',content_type = 'application/json') as s:
                s.write('{ "Key" : "Value" }')
            db._synch()
            meta = db.head('oliver')
            self.assertEqual('application/json',meta.content_type)
    
    def test_should_be_able_to_get_content_type_from_read_stream(self):
        key = 'oliver'
        content_type = 'application/json'
        
        with self.instance() as db:
            with db.write_stream(key,content_type = content_type) as s:
                s.write('{ "Key" : "Value" }')
            db._synch()
            with db.read_stream(key) as rs:
                self.assertEqual(content_type,rs.content_type)
            

class OliverImplementationTests(BaseOliverTests,unittest2.TestCase):
    
    def test_delete_and_synch_writes_to_file(self):
        with self.instance() as db:
            db['Oliver'] = self.random_bytes(100)
            del db['Oliver']
            db._synch()
            self.assertTrue(os.stat(db._delete_fn).st_size)
    
    def test_instance_count_is_correct(self):
        db1 = self.instance()
        db2 = self.instance()
        self.assertEqual(2,db1._instance_count())
        self.assertEqual(2,db2._instance_count())
    
    def test_can_compact_empty_file(self):
        with self.instance() as db:
            db.compact()
            self.assertTrue(os.stat(db._data_fn).st_size)